# -*- coding: utf-8 -*-
import json
import yaml
import socket
import logging
import asyncio
import threading
import tornado.web
import tornado.ioloop
import tornado.httpserver
from tornado.httputil import HTTPHeaders

import os, sys
sys.path.append(os.getcwd())
from app.app_core import AppCore


class GlobalData:
    def __init__(self):
        self.app_core = None
        self.is_gateway = False

    def init_global_data(self):
        try:
            self.app_core = AppCore()
            return True
        except Exception as e:
            logging.error(str(e))
            return False


g = GlobalData()


def gateway_web_service(prev_request, output):

    result = {}
    try:
        result['value'] = g.app_core.get_web_file(prev_request)
        result['status'] = 'ok'
    except Exception as e:
        logging.error(str(e))
        result['value'] = str(e)
        result['status'] = 'err'

    output['result'] = result
    output['finish'] = True


def gateway_data_service(prev_request, output):

    result = {}
    try:
        result['value'] = g.app_core.forward_api(prev_request)
        result['status'] = 'ok'
    except Exception as e:
        logging.error(str(e))
        result['value'] = str(e)
        result['status'] = 'err'

    output['result'] = result
    output['finish'] = True


class GatewayApi(tornado.web.RequestHandler):

    async def get(self, *args, **kwargs):
        output = {
            'result': {},
            'finish': False
        }
        thread = threading.Thread(
            target=gateway_web_service,
            args=(self.request, output)
        )
        thread.setDaemon(True)
        thread.start()

        while not output['finish']:
            await asyncio.sleep(0.01)

        result = output['result']

        if result['status'] == 'ok':
            new_access_token = result['value']['new_access_token']
            new_refresh_token = result['value']['new_refresh_token']
            response = result['value']['response']
            if new_access_token is not None and new_refresh_token is not None:
                self.set_cookie('access_token', new_access_token)
                self.set_cookie('refresh_token', new_refresh_token)

            self.set_status(response.status_code)
            self._headers = HTTPHeaders(response.headers)  # 注意：需要强制类型转换
            self.set_header("Access-Control-Allow-Origin", "*")  # 允许跨域访问
            self.write(response.content)
        else:
            self.set_status(500)
            self.write(result['value'])

    async def post(self, *args, **kwargs):
        output = {
            'result': {},
            'finish': False
        }
        thread = threading.Thread(
            target=gateway_data_service,
            args=(self.request, output)
        )
        thread.setDaemon(True)
        thread.start()

        while not output['finish']:
            await asyncio.sleep(0.01)

        result = output['result']

        if result['status'] == 'ok':
            new_access_token = result['value'].get('new_access_token')  # 用 get 而不是 [] 来取值，因为可能不存在
            new_refresh_token = result['value'].get('new_refresh_token')
            response = result['value']['response']

            if new_access_token is not None and new_refresh_token is not None:
                self.set_cookie('access_token', new_access_token)
                self.set_cookie('refresh_token', new_refresh_token)

            self.set_status(response.status_code)
            try:
                json.loads(response.text, encoding='utf-8')
            except:
                self._headers = HTTPHeaders(response.headers)  # 注意：需要强制类型转换
            self.set_header('Access-Control-Allow-Credentials', 'true')
            self.set_header('Access-Control-Allow-Origin', self.request.headers.get('Origin'))
            self.write(response.content)
        else:
            self.set_status(500)
            self.write(result['value'])

    async def options(self, *args, **kwargs):
        # 允许跨域
        self.set_status(204)
        self.set_header('Access-Control-Allow-Credentials', 'true')
        self.set_header('Access-Control-Allow-Origin', self.request.headers.get('Origin'))
        self.set_header("Access-Control-Allow-Headers", "content-type")
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')


def data_service(input, output, http_request):
    if not isinstance(input, dict):
        result = {
            'value': 'HTTP请求体不是JSON格式！',
            'status': 'err'
        }
        output['result'] = result
        output['finish'] = True
        return

    action = input.get('action')
    kwargs = input.get('args')
    if kwargs is None:
        kwargs = {}
    kwargs['http_request'] = http_request
    
    if action is None:
        result = {
            'value': 'HTTP请求体中未携带action！',
            'status': 'err'
        }
        output['result'] = result
        output['finish'] = True
        return

    public_actions = getattr(AppCore, 'public_actions', None)
    if public_actions is not None:
        if action not in public_actions:
            result = {
                'value': 'Action: {} 禁止访问！'.format(action),
                'status': 'err'
            }
            output['result'] = result
            output['finish'] = True
            return

    model = g.app_core
    action_obj = getattr(model, action, None)
    if action_obj is None:
        result = {
            'value': 'Action: {} 未定义！'.format(action),
            'status': 'err'
        }
        output['result'] = result
        output['finish'] = True
        return

    result = {}
    try:
        result['value'] = action_obj(**kwargs) if kwargs else action_obj()  # kwargs为None或{}时，不带参数调用
        result['status'] = 'ok'
    except Exception as e:
        logging.error(str(e))
        result['value'] = str(e)
        result['status'] = 'err'

    output['result'] = result
    output['finish'] = True
    

def stream_service(input, output, http_request):

    action = input.get('action')
    arg = input.get('arg')

    public_actions = getattr(AppCore, 'public_actions', None)
    if public_actions is not None:
        if action not in public_actions:
            result = {
                'value': 'Action: {} 禁止访问！'.format(action),
                'status': 'err'
            }
            output['result'] = result
            output['finish'] = True
            return

    model = g.app_core
    action_obj = getattr(model, action, None)
    if action_obj is None:
        result = {
            'value': 'Action: {} 未定义！'.format(action),
            'status': 'err'
        }
        output['result'] = result
        output['finish'] = True
        return

    result = {}
    try:
        result['value'] = action_obj(arg, http_request)
        result['status'] = 'ok'
    except Exception as e:
        logging.error(str(e))
        result['value'] = str(e)
        result['status'] = 'err'

    output['result'] = result
    output['finish'] = True


class DataApi(tornado.web.RequestHandler):

    async def post(self, *args, **kwargs):

        try:
            input = json.loads(str(self.request.body, encoding='utf-8'))
        except Exception as e:
            logging.error(str(e))
            result = {
                'status': 'err',
                'value': 'HTTP请求体错误！' + str(e)
            }
            res = json.dumps(result, ensure_ascii=False)
            self.set_header('Access-Control-Allow-Origin', '*')  # 允许跨域访问
            self.write(res)
            return 
            
        output = {
            'result': {},
            'finish': False
        }
        thread = threading.Thread(
            target=data_service,
            args=(input, output, self.request)
        )
        thread.setDaemon(True)
        thread.start()

        while not output['finish']:
            await asyncio.sleep(0.01)

        result = output['result']

        if isinstance(result['value'], bytes):
            res = result['value']
        else:
            # 特例，针对gateway上的 login 和 logout
            if g.is_gateway and result['status'] == 'ok':
                self.set_cookie('access_token', result['value']['access_token'])
                self.set_cookie('refresh_token', result['value']['refresh_token'])
            
            try:
                res = json.dumps(result, ensure_ascii=False)
            except Exception as e:
                logging.error(str(e))
                result = {
                    'status': 'err',
                    'value': 'API调用返回结果序列化失败！' + str(e)
                }
                res = json.dumps(result, ensure_ascii=False)

        self.set_header('Access-Control-Allow-Origin', '*')  # 允许跨域访问
        self.write(res)

    async def options(self, *args, **kwargs):
        self.set_status(204)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'content-type')
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')


class StreamApi(tornado.web.RequestHandler):

    async def post(self, action, *args, **kwargs):

        input = {
            'action': action,
            'arg': self.request.body,
        }
        output = {
            'result': {},
            'finish': False
        }
        thread = threading.Thread(
            target=stream_service,
            args=(input, output, self.request)
        )
        thread.setDaemon(True)
        thread.start()

        while not output['finish']:
            await asyncio.sleep(0.01)

        result = output['result']

        if isinstance(result['value'], bytes):
            res = result['value']
        else:
            try:
                res = json.dumps(result, ensure_ascii=False)
            except Exception as e:
                logging.error(str(e))
                result = {
                    'status': 'err',
                    'value': 'API调用返回结果序列化失败！' + str(e)
                }
                res = json.dumps(result, ensure_ascii=False)

        self.set_header('Access-Control-Allow-Origin', '*')  # 允许跨域访问
        self.write(res)

    async def options(self, *args, **kwargs):
        self.set_status(204)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'content-type')
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')


class FileApi(tornado.web.RequestHandler):

    async def post(self, action, *args, **kwargs):
        
        try:
            file_obj = self.request.files.get(action)[0]
            file_body = file_obj.body
        except Exception as e:
            logging.error(str(e))
            result = {
                'status': 'err',
                'value': 'HTTP文件上传请求体错误！' + str(e)
            }
            res = json.dumps(result, ensure_ascii=False)
            self.set_header('Access-Control-Allow-Credentials', 'true')
            self.set_header('Access-Control-Allow-Origin', self.request.headers.get('Origin'))
            self.write(res)
            return 

        input = {
            'action': action,
            'arg': file_body,
        }
        output = {
            'result': {},
            'finish': False
        }
        thread = threading.Thread(
            target=stream_service,
            args=(input, output, self.request)
        )
        thread.setDaemon(True)
        thread.start()

        while not output['finish']:
            await asyncio.sleep(0.01)

        result = output['result']

        if isinstance(result['value'], bytes):
            res = result['value']
        else:
            try:
                res = json.dumps(result, ensure_ascii=False)
            except Exception as e:
                logging.error(str(e))
                result = {
                    'status': 'err',
                    'value': 'API调用返回结果序列化失败！' + str(e)
                }
                res = json.dumps(result, ensure_ascii=False)

        self.set_header('Access-Control-Allow-Credentials', 'true')
        self.set_header('Access-Control-Allow-Origin', self.request.headers.get('Origin'))
        self.write(res)

    async def options(self, *args, **kwargs):
        self.set_status(204)
        self.set_header('Access-Control-Allow-Credentials', 'true')
        self.set_header('Access-Control-Allow-Origin', self.request.headers.get('Origin'))
        self.set_header('Access-Control-Allow-Headers', 'content-type')
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')


class HealthChecker(tornado.web.RequestHandler):
    def get(self, *args, **kwargs):
        self.write('{"description": "Micro-service Discovery Client", "status": "UP"}')


def get_local_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()

    return ip


def start():
    app_profile = os.environ.get('APP_PROFILE', 'dev').lower()
    log_level = logging.DEBUG if app_profile == 'dev' else logging.ERROR
    logging.basicConfig(level=log_level, format='%(asctime)s - %(levelname)s - %(message)s')

    try:
        with open('./application.yml', 'r') as f:
            yml = yaml.load(f, Loader=yaml.SafeLoader)
    except:
        logging.error('服务配置文件application.yml不存在！')
        return
    
    try:
        ename = yml['service']['ename']
    except:
        logging.error('未指定服务英文名！请在application.yml文件中编辑修改...')
        return

    try:
        cname = yml['service']['cname']
    except:
        cname = ename.upper()

    try:
        port = yml['service']['port']
    except:
        logging.error('未指定服务端口号！缺省使用80端口。')
        port = 80

    if not g.init_global_data():
        logging.error('微服务： {}/{} 初始化失败！'.format(cname, ename))
        return

    is_gateway = False
    try:
        is_gateway = yml['gateway']['is_gateway']
    except:
        pass

    if is_gateway:
        g.is_gateway = True
        handlers = [
            (r'/management/health', HealthChecker),
            (r'/api/data', DataApi),
            (r'/(.*)', GatewayApi),
        ]
    else:
        handlers = [
            (r'/management/health', HealthChecker),
            (r'/api/data', DataApi),
            (r'/api/stream/(\w+)', StreamApi),
            (r'/api/file/(\w+)', FileApi),
            (r'/(.*)', tornado.web.StaticFileHandler, {'path': 'webapp/www', 'default_filename': 'index.html'}),
        ]

    app = tornado.web.Application(
        handlers=handlers,
        debug=('dev' == app_profile)
    )
    http_server = tornado.httpserver.HTTPServer(app, max_buffer_size=1000*1024*1024)
    http_server.listen(port)

    logging.critical('##################################################')
    logging.critical('    微服务： {}/{} started ...'.format(cname, ename))
    logging.critical('    Listening at: {}:{}'.format(get_local_ip(), port))
    logging.critical('    App profile: {}'.format(app_profile))
    logging.critical('##################################################')
    tornado.ioloop.IOLoop.current().start()


if __name__ == '__main__':
    start()
